/* vim: set sw=4 sts=4 et foldmethod=syntax : */

/*
 * Copyright (c) 2006 Ciaran McCreesh <ciaranm@ciaranm.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the library nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef LIBWRAPITER_GUARD_LIBWRAPITER_LIBWRAPITER_FORWARD_ITERATOR_HH
#define LIBWRAPITER_GUARD_LIBWRAPITER_LIBWRAPITER_FORWARD_ITERATOR_HH 1

#include <libwrapiter/libwrapiter_fwd.hh>
#include <iterator>
#include <utility>

/** \file
 * Header for the ForwardIterator class and its implementation.
 */

namespace libwrapiter
{
    /**
     * The underlying iterator wrapper for a ForwardIterator.
     */
    template <typename Type_>
    class ForwardIteratorBase
    {
        private:
            ///\name Disallow copying and assignment
            ///\{

            ForwardIteratorBase(const ForwardIteratorBase &);
            ForwardIteratorBase operator= (const ForwardIteratorBase &);

            ///\}

        protected:
            ///\name Basic operations
            ///\{

            ForwardIteratorBase()
            {
            }

            ///\}

        public:
            ///\name Basic operations
            ///\{

            /**
             * Copy ourself.
             *
             * This can't be a regular copy constructor since copying removes
             * subclassing.
             */
            virtual ForwardIteratorBase<Type_> * clone() = 0;

            /**
             * Destructor.
             */
            virtual ~ForwardIteratorBase()
            {
            }

            ///\}

            ///\name Comparison operators
            ///\{

            /**
             * Equality operator.
             */
            virtual bool operator== (const ForwardIteratorBase<Type_> & other) const = 0;

            /**
             * Inequality operator.
             */
            virtual bool operator!= (const ForwardIteratorBase<Type_> & other) const = 0;

            ///\}

            ///\name Dereference operators
            ///\{

            /**
             * Dereference operator.
             */
            virtual Type_ & operator* () const = 0;

            /**
             * Dereference operator.
             */
            virtual Type_ * operator-> () const = 0;

            ///\}

            ///\name Increment operators
            ///\{

            /**
             * Increment, used by both operator++ methods.
             */
            virtual void increment() = 0;

            ///\}
    };

    /**
     * The implementation of a ForwardIteratorBase for a particular iterator
     * type.
     */
    template <typename Type_, typename Iter_>
    class ForwardIteratorBaseImpl :
        public ForwardIteratorBase<Type_>
    {
        private:
            /// Our underlying iterator.
            Iter_ _iter;

        public:
            ///\name Basic operations
            ///\{

            /**
             * Constructor, from an iterator.
             */
            ForwardIteratorBaseImpl(const Iter_ & i) :
                _iter(i)
            {
            }

            ///\}

            virtual ForwardIteratorBase<Type_> * clone()
            {
                return new ForwardIteratorBaseImpl<Type_, Iter_>(_iter);
            }

            virtual bool operator== (const ForwardIteratorBase<Type_> & other) const
            {
                return _iter == static_cast<const ForwardIteratorBaseImpl<Type_, Iter_> *>(&other)->_iter;
            }

            virtual bool operator!= (const ForwardIteratorBase<Type_> & other) const
            {
                return _iter != static_cast<const ForwardIteratorBaseImpl<Type_, Iter_> *>(&other)->_iter;
            }

            virtual Type_ & operator* () const
            {
                return *_iter;
            }

            virtual Type_ * operator-> () const
            {
                return &*_iter;
            }

            virtual void increment()
            {
                ++_iter;
            }
    };

    /**
     * A generic forward iterator that points to an item of type Type_.
     *
     * The template is parameterised by OurType_, which should be a unique
     * class (usually the owning class) for each iterator. This is to
     * prevent comparing, say, a ForwardIterator<int> implemented in terms of a
     * set iterator with a ForwardIterator<int> implemented in terms of a list
     * iterator. The Type_ parameter is the type of the underlying item (for
     * const iterators, it should be a const type).
     */
    template <typename OurType_, typename Type_>
    class ForwardIterator :
        public std::iterator<std::forward_iterator_tag, Type_>
    {
        private:
            /// Our underlying handler.
            ForwardIteratorBase<Type_> * _base;

        public:
            ///\name Basic operations
            ///\{

            /**
             * Constructor, from an iterator.
             */
            template <typename Iter_>
            explicit ForwardIterator(const Iter_ & iter) :
                _base(new ForwardIteratorBaseImpl<Type_, Iter_>(iter))
            {
            }

            /**
             * Copy constructor.
             */
            ForwardIterator(const ForwardIterator<OurType_, Type_> & other) :
                _base(other._base->clone())
            {
            }

            /**
             * Assignment operator.
             */
            const ForwardIterator & operator= (const ForwardIterator<OurType_, Type_> & other)
            {
                if (this != &other)
                {
                    ForwardIteratorBase<Type_> * _new_base(other._base->clone());
                    std::swap(_base, _new_base);
                    delete _new_base;
                }
                return *this;
            }

            /**
             * Destructor.
             */
            ~ForwardIterator()
            {
                delete _base;
            }

            ///\}

            ///\name Comparison operators
            ///\{

            /**
             * Equality operator.
             */
            bool operator== (const ForwardIterator<OurType_, Type_> & other) const
            {
                return *_base == *other._base;
            }

            /**
             * Inequality operator.
             */
            bool operator!= (const ForwardIterator<OurType_, Type_> & other) const
            {
                return *_base != *other._base;
            }

            ///\}

            ///\name Dereference operators
            ///\{

            /**
             * Dereference operator.
             */
            Type_ & operator* () const
            {
                return **_base;
            }

            /**
             * Dereference operator.
             */
            Type_ * operator-> () const
            {
                return _base->operator->();
            }

            ///\}

            ///\name Increment operators
            ///\{

            /**
             * Preincrement.
             */
            ForwardIterator<OurType_, Type_> & operator++ ()
            {
                _base->increment();
                return *this;
            }

            /**
             * Postincrement.
             */
            ForwardIterator<OurType_, Type_> operator++ (int)
            {
                ForwardIterator<OurType_, Type_> result(*this);
                _base->increment();
                return result;
            }

            ///\}
    };
}

#endif
