/* vim: set sw=4 sts=4 et foldmethod=syntax : */

/*
 * Copyright (c) 2005, 2006 Ciaran McCreesh <ciaranm@ciaranm.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the library nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "test_framework.hh"
#include <algorithm>
#include <iostream>
#include <unistd.h>

/** \file
 * Implementation for test framework classes.
 */

using namespace test;

#ifndef DOXYGEN
namespace
{
    std::string exception_to_debug_string(const std::exception & e)
    {
        return e.what() + std::string(" (no further information)");
    }

    class DebugStringHolder
    {
        private:
            std::string (*_f) (const std::exception &);

            DebugStringHolder() :
                _f(&exception_to_debug_string)
            {
            }

        public:
            static DebugStringHolder * get_instance()
            {
                static DebugStringHolder _instance;
                return &_instance;
            }

            void set(std::string (*f) (const std::exception &))
            {
                _f = f;
            }

            std::string (*get() const) (const std::exception &)
            {
                return _f;
            }
    };
}

void
test::set_exception_to_debug_string(std::string (*f) (const std::exception &))
{
    DebugStringHolder::get_instance()->set(f);
}

std::list<std::string>
TestMessageSuffix::_suffixes;

std::string
TestMessageSuffix::suffixes()
{
    std::string result;
    std::list<std::string>::const_iterator i(_suffixes.begin()), end(_suffixes.end());
    while (i != end)
    {
        result += *i++;
        if (end != i)
            result += ", ";
    }
    return result;
}

TestMessageSuffix::TestMessageSuffix(const std::string & s, bool write)
{
    _suffixes.push_back(s);
    if (write)
        std::cout << "[" << s << "]" << std::flush;
}

struct TestCase::Impl
{
    const std::string name;

    Impl(const std::string & the_name) :
        name(the_name)
    {
    }
};

TestCase::TestCase(const std::string & n) :
    _impl(new Impl(n))
{
    TestCaseList::register_test_case(this);
}

TestCase::~TestCase()
{
}

void
TestCase::check(const char * const function, const char * const file,
        const long line, bool was_ok, const std::string & message) const
{
    std::cout << "." << std::flush;
    if (! was_ok)
        throw TestFailedException(function, file, line, message);
}

void
TestCase::call_run()
{
    try
    {
        run();
    }
    catch (TestFailedException)
    {
        throw;
    }
    catch (std::exception &e)
    {
        throw TestFailedException(__PRETTY_FUNCTION__, __FILE__, __LINE__,
                "Test threw unexpected exception " + (DebugStringHolder::get_instance()->get())(e));
    }
    catch (...)
    {
        throw TestFailedException(__PRETTY_FUNCTION__, __FILE__, __LINE__,
                "Test threw unexpected unknown exception");
    }
}

std::string
TestCase::name() const
{
    return _impl->name;
}

namespace
{
    std::string
    line_to_string(const long line)
    {
        std::stringstream s;
        s << line;
        return s.str();
    }
}

TestFailedException::TestFailedException(const char * const function, const char * const file,
        const long line, const std::string & message) throw () :
    _message(std::string(file) + ":" + line_to_string(line) + ": in " +
            std::string(function) + ": " + message + (
                TestMessageSuffix::suffixes().empty() ? std::string("") : " [context: " +
                TestMessageSuffix::suffixes() + "]"))
{
}

TestFailedException::~TestFailedException() throw ()
{
}

TestCaseList::TestCaseList()
{
}

TestCaseList::~TestCaseList()
{
}

class RunTest
{
    private:
        bool * const _had_a_failure;

    public:
        RunTest(bool * had_a_failure) :
            _had_a_failure(had_a_failure)
        {
        }

        void operator() (TestCase * test_case) const;
};

void
RunTest::operator() (TestCase * test_case) const
{
    bool had_local_failure(false);

    std::cout << "* \"" << test_case->name() << "\": " << std::flush;

    for (int repeat = 0 ; repeat < 2 ; ++repeat)
    {
        if (test_case->skip())
        {
            std::cout << "(skip)" << std::endl;
            break;
        }

        if (0 != repeat)
            std::cout << "  (repeat): " << std::flush;

        try
        {
            if (TestCaseList::use_alarm)
                alarm(test_case->max_run_time());
            test_case->call_run();
            if (TestCaseList::use_alarm)
                alarm(0);
        }
        catch (std::exception &e)
        {
            std::cout << "!{" << std::endl << (DebugStringHolder::get_instance()->get())(e) <<
                std::endl << "  } " << std::flush;
            had_local_failure = true;
            *_had_a_failure = true;
        }
        catch (...)
        {
            std::cout << "!{Unknown exception type} ";
            had_local_failure = true;
            *_had_a_failure = true;
        }

        if (had_local_failure)
            std::cout << " NOT OK";
        else
            std::cout << " OK";

        std::cout << std::endl;

        if (! test_case->repeatable())
            break;
    }
}

bool
TestCaseList::run_tests()
{
    bool had_a_failure(_get_test_case_list().empty());

    std::for_each(
            TestCaseList::_get_test_case_list().begin(),
            TestCaseList::_get_test_case_list().end(),
            RunTest(&had_a_failure));

    return ! had_a_failure;
}

bool
TestCaseList::use_alarm(true);

#endif

