/* vim: set sw=4 sts=4 et foldmethod=syntax : */

/*
 * Copyright (c) 2005, 2006 Ciaran McCreesh <ciaranm@ciaranm.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the library nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GUARD_TEST_TEST_FRAMEWORK_HH
#define GUARD_TEST_TEST_FRAMEWORK_HH 1

#include <list>
#include <memory>
#include <sstream>
#include <string>

/** \file
 * Test framework class definitions.
 *
 * \ingroup Test
 */

namespace test
{
    /**
     * RAII suffix marker for TestCase context.
     *
     * \ingroup Test
     */
    class TestMessageSuffix
    {
        private:
            static std::list<std::string> _suffixes;

        public:
            /**
             * Constructor.
             */
            explicit TestMessageSuffix(const std::string & s, bool write = false);

            /**
             * Destructor.
             */
            ~TestMessageSuffix()
            {
                _suffixes.pop_back();
            }

            /**
             * Our suffixes.
             */
            static std::string suffixes();
    };

    /**
     * Base TestCase class.
     *
     * \ingroup Test
     */
    class TestCase
    {
        private:
            struct Impl;
            std::auto_ptr<Impl> _impl;

        protected:
            /**
             * Check that a given assertion is true.
             */
            void check(const char * const function, const char * const file,
                    const long line, bool was_ok, const std::string & message) const;

            /**
             * Run the tests (override this in descendents).
             */
            virtual void run() = 0;

        public:
            /**
             * Constructor.
             */
            TestCase(const std::string & name);

            /**
             * Destructor.
             */
            virtual ~TestCase();

            /**
             * Call our run() function, and normalise exceptions.
             */
            void call_run();

            /**
             * Return our name.
             */
            std::string name() const;

            /**
             * Are we repeatable? Override if not.
             */
            virtual bool repeatable() const
            {
                return true;
            }

            /**
             * Should we be skipped?
             */
            virtual bool skip() const
            {
                return false;
            }

            /**
             * After how many seconds should we timeout?
             */
            virtual unsigned max_run_time() const
            {
                return 30;
            }
    };

    /**
     * Thrown if a TestCase failed.
     *
     * \ingroup Test
     * \ingroup grpexceptions
     */
    class TestFailedException : public std::exception
    {
        private:
            const std::string _message;

        public:
            /**
             * Constructor.
             */
            TestFailedException(const char * const function, const char * const file,
                    const long line, const std::string & message) throw ();

            /**
             * Destructor.
             */
            virtual ~TestFailedException() throw ();

            /**
             * Description.
             */
            const char * what() const throw ()
            {
                return _message.c_str();
            }
    };

    /**
     * A list of TestCase instances.
     *
     * \ingroup Test
     */
    class TestCaseList
    {
        private:
            TestCaseList();
            ~TestCaseList();

            static std::list<TestCase *> & _get_test_case_list()
            {
                static std::list<TestCase *> l;
                return l;
            }

        public:
            /**
             * Register a TestCase instance.
             */
            inline static void register_test_case(TestCase * const test_case)
            {
                _get_test_case_list().push_back(test_case);
            }

            /**
             * Run all tests.
             */
            static bool run_tests();

            /**
             * Should we use alarm?
             */
            static bool use_alarm;
    };

    /**
     * Change the function used to get a string description of an exception.
     */
    void set_exception_to_debug_string(std::string (*) (const std::exception &));
}

/**
 * Check that a == b.
 */
#define TEST_CHECK_EQUAL(a, b) \
    do { \
        std::stringstream a_str, b_str; \
        a_str << a; \
        b_str << b; \
        check(__PRETTY_FUNCTION__, __FILE__, __LINE__, a == b, \
                "Expected '" #a "' to equal '" + b_str.str() + \
                "' but got '" + a_str.str() + "'"); \
    } while (false)

/**
 * Check that stringify(a) == stringify(b).
 */
#define TEST_CHECK_STRINGIFY_EQUAL(a, b) \
    do { \
        std::stringstream a_str, b_str; \
        a_str << a; \
        b_str << b; \
        check(__PRETTY_FUNCTION__, __FILE__, __LINE__, a_str.str() == b_str.str(), \
                "Expected '" #a "' to equal '" + b_str.str() + \
                "' but got '" + a_str.str() + "'"); \
    } while (false)

/**
 * Check that a is true.
 */
#define TEST_CHECK(a) \
    do { \
        check(__PRETTY_FUNCTION__, __FILE__, __LINE__, a, \
                "Check '" #a "' failed"); \
    } while (false)

/**
 * Check that a throws an exception of type b.
 */
#define TEST_CHECK_THROWS(a, b) \
    do { \
        try { \
            a; \
            check(__PRETTY_FUNCTION__, __FILE__, __LINE__, false, \
                    "Expected exception of type '" #b "' not thrown"); \
        } catch (b &) { \
            TEST_CHECK(true); \
        } \
    } while (false)

#endif

