/* vim: set sw=4 sts=4 et foldmethod=syntax : */

/*
 * Copyright (c) 2006 Ciaran McCreesh <ciaranm@ciaranm.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the library nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "test_framework.hh"
#include "test_runner.hh"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <signal.h>
#if defined(__GLIBC__)
#include <execinfo.h>
#endif
#include <unistd.h>
#include <sys/types.h>

/** \file
 * Implementation of the default test runner.
 *
 * \ingroup Test
 */

using namespace test;

namespace
{
    void do_backtrace()
    {
#if defined(__GLIBC__)
        void * bt[50];
        size_t sz = backtrace(bt, 50);
        char * * symbols = backtrace_symbols(bt, sz);

        std::cerr << "Stack dump:" << std::endl;
        for (unsigned n(0) ; n < sz ; ++n)
            std::cerr << "  * " << symbols[n] << std::endl;

        std::free(symbols);
#endif
    }

#if ((__GNUC__ > 3) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
    void timeout_handler(int) __attribute__((noreturn));
    void segfault_handler(int) __attribute__((noreturn));
#endif

    void timeout_handler(int)
    {
        std::cerr << std::endl << "Test aborted due to timeout!" << std::endl;
        do_backtrace();
        std::exit(EXIT_FAILURE);
    }

    void segfault_handler(int)
    {
        std::cerr << std::endl << "Test aborted due to segmentation fault!" << std::endl;
        do_backtrace();
        std::exit(EXIT_FAILURE);
    }
}

namespace
{
    std::string
    pid_t_to_string(const pid_t p)
    {
        std::stringstream s;
        s << p;
        return s.str();
    }
}


int
main(int, char * argv[])
{
    {
        std::ifstream ppid(("/proc/" + pid_t_to_string(getppid()) + "/cmdline").c_str());
        if (ppid)
        {
            std::string cmd;
            std::getline(ppid, cmd, '\0');
            std::string::size_type slash_pos(cmd.rfind('/'));
            if (std::string::npos != slash_pos)
                cmd.erase(0, slash_pos);
            if (cmd != "gdb")
            {
                signal(SIGALRM, &timeout_handler);
                signal(SIGSEGV, &segfault_handler);
            }
            else
                TestCaseList::use_alarm = false;
        }
    }

    std::cout << "Test program " << argv[0] << ":" << std::endl;
    return TestCaseList::run_tests() ? EXIT_SUCCESS : EXIT_FAILURE;
}

